/* Copyright 2009, UCAR/Unidata and OPeNDAP, Inc.
   See the COPYRIGHT file for more information. */

/* -*- mode: c++; c-basic-offset:4 -*-

/* IDL-OPENDAP interface.
// Copyright (c) 2006 OPeNDAP, Inc.
// Author: Patrick West <mrpatrickwest@gmail.com> UCAR/HAO
//         Dan Holloway <d.holloway@opendap.org>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.

// Please read the full copyright statement in the file COPYING.
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#ifndef WIN32
#include <strings.h>
#endif

/* OC include*/
#include "ocinternal.h"
#include "ocdata.h"
#include "occontent.h"
#include "i.h"
#include "ocdebug.h"

/* Forward*/
static int consistent(OCnode* ocnode);
static int findfield(OCnode* prevnode, OCnode* node);
static OCcontent* Ilocatecontent(OCstate* state, OCnode* node);
static OCcontent* Ilocatecontent0(OCstate* state, OCcontent* parentcontent, OCnode* node);
static OCcontent* Imakecontent(OCstate* conn, OCnode* node);
static int Isetfieldindex(OCstate* conn, OCcontent* content, size_t fieldindex);

/*
opendap_connect()
 Establish connection and return connect object and DDS object.
    argv[0] - (Input ) url byte stream (null terminated).
    argv[1] - (Input ) Constraint expression byte stream (null terminated).
    argv[2] - (Output) Named IDL variable to recieve a pointer
    argv[3] - (Input ) ACCEPT_DEFLATE flag.
    argv[4] - (Input ) A flag indicating what kind of data request is:
    data=0, dds=1, das=2, fulldata=3, fulldds=4, fulldas=5
    returns -          Status (1=ok, 0=fail).
*/

int
opendap_connect(char* url, char* expr, OCnode** nodep, int deflate, int connect_mode)
{
    int stat;
    char *ext;
    char *fullurl = 0;
    OCnode* node;
    OCstate* state;
    OCcontent* rootcontent;

    switch(connect_mode) {
    case 1:
    case 4: ext = ".dds"; break;
    case 2:
    case 5: ext = ".das"; break;
    case 0:
    case 3: ext = ".dods"; break;
    default: return THROW(OC_EINVAL);
    }

    if(connect_mode == 1 && strlen( expr ) != 0 ) {
      int fullsize = strlen( url ) + strlen( ext ) + strlen( expr );
      fullurl = (char *)ocmalloc( fullsize + 2 );
      sprintf( fullurl, "%s?%s", url, expr );
    } else {
      int fullsize;
      if(strlen( expr ) != 0 ) {
        fullsize = strlen( url ) + strlen( ext ) + strlen( expr );
        fullurl = (char *)ocmalloc( fullsize + 2 );
        sprintf( fullurl, "%s?%s", url, expr );
      } else {
        fullsize = strlen( url ) + strlen( ext );
        fullurl = (char *)ocmalloc( fullsize + 1 );
        sprintf( fullurl, "%s", url);
      }
    }

    stat = oc_open(&state);
    if(stat != OC_NOERR) return THROW(stat);

    switch (connect_mode) {
    case 1: /* dds*/
    case 4: /* fulldds*/
	stat = oc_fetchdds(state,fullurl);
        if(stat != OC_NOERR) return THROW(stat);
	if(!consistent(state->schema.dds.tree))
	    return THROW(OC_EINVAL);
        node = state->schema.dds.tree;
	*nodep = node;
	break;

    case 2: /* das*/
    case 5: /* fulldas*/
	stat = oc_fetchdas(state,fullurl);
        if(stat != OC_NOERR) return THROW(stat);
        node = state->schema.das.tree;
	*nodep = node;
	break;

    case 0: /* data*/
    case 3: /* fulldata*/
	stat = oc_fetchdatadds(state,fullurl);
        if(stat != OC_NOERR) return THROW(stat);
	if(!consistent(state->schema.datadds.tree))
	    return THROW(OC_EINVAL);
        node = state->schema.datadds.tree;
	*nodep = node;
	/* Connect up the content*/
	rootcontent = Imakecontent(state,node);
        stat = oc_rootcontent(state,rootcontent);
        if(stat != OC_NOERR) return THROW(stat);
	break;

    default: return THROW(OC_EINVAL);
    }

    /* free allocated memory*/
    free( fullurl );

    return THROW(OC_NOERR);
}

static OCcontent*
Imakecontent(OCstate* conn, OCnode* node)
{
    OCcontent* oc = oc_newcontent(conn);
    node->public = (void*)oc; /* connect content to the node*/
    oc->node = node;
    return oc;
}

static int
Isetfieldindex(OCstate* conn, OCcontent* content, size_t fieldindex)
{
    Icontent * ic = (Icontent*)content->public;
    if(ic == NULL) {
        ic = (Icontent*)ocmalloc(sizeof(Icontent*)); /* ocmalloc zeros*/
        content->public = (void*)ic;
    }
    ic->fieldindex = fieldindex;
    return THROW(OC_NOERR);
}

/*
opendap_node_info()
 Get information about a node.
    argv[0] - (Input ) node pointer.
    argv[1] - (Output) Named IDL variable to receive node name
    argv[2] - (Output) Named IDL variable to receive node type
    argv[3] - (Output) Named IDL variable to receive number of sub nodes
    argv[4] - (Output) Named IDL variable to receive whether node has content
    argv[5] - (Output) Named IDL variable to receive whether node has attrs
    argv[6] - (Output) Named IDL variable to receive node content ptr
*/

/*
There is a problem with opendap_node_info in that it asserts
to provide access to the content of any given node (not just
the root node as in occontent.c. In order to support this
(for IDL only) we need to construct the path to this node
and apply it to the content.
*/

int
opendap_node_info(OCnode* node, char** namep, OCtype* typep, int* nsubnodesp,
                  int* hascontentp, int* hasattrsp, OCcontent** contentp)
{
    int stat = OC_NOERR;

    *namep = strdup(node->name);
    *typep = node->octype;
    *nsubnodesp = oclistlength(node->subnodes);
    *hasattrsp = 0;

    *hascontentp = 1;

    *contentp = Ilocatecontent(node->owner,node);

    return THROW(stat);
}

/*
 opendap_get_subnode()
 Get the specified sub node of the specified node
    argv[0] - (Input ) node pointer.
    argv[1] - (Input ) number of subnode to get
    argv[2] - (Output) Named IDL variable to receive subnode
    argv[3] - (Output) Named IDL variable to receive subnode name
*/

int
opendap_get_subnode(OCnode* node, int isubnode, OCnode** subnodep, char** namep)
{
    OCnode* subnode;
    OCstate* state;

    state = node->owner;
    if(isubnode < 0 || isubnode > oclistlength(node->subnodes)) return THROW(OC_EINVAL);
    subnode = (OCnode*)oclistget(node->subnodes,isubnode);
    *namep = strdup(subnode->name);
    *subnodep = subnode;
    return THROW(OC_NOERR);
}

/*
opendap_get_int32()
Get the value of the node as int32
     argv[0] - (Input ) content pointer.
     argv[1] - (Output) Named IDL variable to receive data value
     argv[2] - (Input ) get_data, is this a dds or a datadds
     returns - Status (1=ok, 0=fail)
*/
int
opendap_get_int32(OCcontent* content, int* ivaluep, int mode)
{
    int stat = OC_NOERR;
    OCnode* node = content->node;
    OCstate* state = node->owner;
    XDR* xdrs = state->dap.xdrs;

    if(node->octype != OC_Primitive || node->array.rank > 0) return THROW(OC_EINVAL);

    if(mode == 0) { /* dds => return schema info*/
	*ivaluep = 0;
    } else if(xdrs) { /* data => return actual data*/
	stat = oc_getcontent(state,content,ivaluep,sizeof(int),0,1);
    } else return THROW(OC_EXDR);

    return THROW(OC_NOERR);
}

/*
opendap_get_array()
Get the value of the node as array
     argv[0] - (Input ) content pointer.
     argv[1] - (Output) Named IDL variable to receive array data
     argv[2] - (Output) Named IDL variable to receive array name
     argv[3] - (Output) Named IDL variable to receive dimension names
     argv[4] - (Input ) get_data, is this a dds(0) or a datadds(1)
     returns - Status (1=ok, 0=fail)
*/

int
opendap_get_array(OCcontent* content, void* memoryp, char** arraynamep,
                  char*** dimnamesp, int mode)
{
    int i;
    OCnode* node = content->node;
    OCstate* state = node->owner;
    XDR* xdrs = state->dap.xdrs;
    char** dimnames;
    size_t n_elements;
    int num_dims = node->array.rank;

    if(node->octype != OC_Primitive) return THROW(OC_EINVAL);

    *arraynamep = strdup(node->name);

    dimnames = (char**)ocmalloc(sizeof(char*)*(num_dims+1)); /* null terminate*/
    for(i=0;i<node->array.rank;i++) {
	OCnode* dim = (OCnode*)oclistget(node->array.dimensions,i);
	if(dim->name == NULL)
	    dimnames[i] = strdup(" ");
	else
	    dimnames[i] = strdup(dim->name);
    }
    dimnames[num_dims] = NULL;
    *dimnamesp = dimnames;

    n_elements = 1;
    for(i=0;i<num_dims;i++) {
	OCnode* dim = (OCnode*)oclistget(node->array.dimensions,i);
        n_elements *= dim->dim.declsize;
    }

    if(mode == 0) { /* dds => return dim count and sizes*/
        /* dds request*/
        int szDims;
        int *size_ptr = 0;
        szDims = (int) (num_dims + 3);
        size_ptr = (int*)ocmalloc(sizeof(int)*szDims);
        size_ptr[0] = (int) num_dims;
        for(i=0;i< num_dims;i++) {
	    OCnode* dim = (OCnode*)oclistget(node->array.dimensions,i);
            size_ptr[i + 1] = dim->dim.declsize;
        }
        size_ptr[num_dims + 1] = (int) node->etype;
        size_ptr[num_dims + 2] = (int) n_elements;
	if(memoryp != NULL) *((void**)memoryp) = (void*) size_ptr;
    } else if(xdrs) { /* return real data*/
	int stat = OC_NOERR;
	size_t typesize = octypesize(node->etype);
	size_t memsize = typesize*n_elements;
	size_t offset, avail;
	char* memory = (char*)ocmalloc(memsize);
	offset = 0;
	avail = memsize;
	for(i=0;i<n_elements;i++) {
	    stat = oc_getcontent(state,content,memory+offset,avail,i,1);
            if(stat != OC_NOERR) return THROW(stat);
	    offset += typesize;
	    avail -= typesize;
        }
	if(memoryp) *(void**)memoryp = (void*)memory;
    } else return THROW(OC_EXDR);
    return THROW(OC_NOERR);
}

/*
opendap_get_struct_info()
Get the information about a structure
     argv[0] - (Input ) content pointer.  ? correct, but not very sensical
     argv[1] - (Output) Named IDL variable to receive num_members
*/
int
opendap_get_struct_info(OCcontent* content, int* nmembersp)
{
    OCnode* node = content->node;

    if(node->octype != OC_Structure && node->octype != OC_Dataset) return THROW(OC_EINVAL);
    *nmembersp = oclistlength(node->subnodes);
    return THROW(OC_NOERR);
}

/*
opendap_get_struct_member()
Get the content for the specified member of a NodeContent_Structure
     argv[0] - (Input ) content pointer.
     argv[1] - (Input ) member number to get
     argv[2] - (Output) Named IDL variable to receive member type
     argv[3] - (Output) Named IDL variable to receive member content ptr
*/
int
opendap_get_struct_member(OCcontent* content, int imember, OCtype* typep, OCcontent** newcontentp)
{
    OCnode* node = content->node;
    OCnode* field;

    if(node->octype != OC_Structure && node->octype != OC_Dataset) return THROW(OC_EINVAL);

    if(imember < 0 || imember >= oclistlength(node->subnodes)) return THROW(OC_EINVALCOORDS);
    field = (OCnode*)oclistget(node->subnodes,imember);
    *typep = field->octype;

    content->index = imember;
    *newcontentp = Ilocatecontent(content->state,field);

    return THROW(OC_NOERR);
}

/*
opendap_get_sequence_info()
   argv[0] - (Input ) content pointer.
   argv[1] - (Output) Named IDL variable to receive num_rows
   argv[2] - (Output) Named IDL variable to receive num_fields
*/

int
opendap_get_sequence_info(OCcontent* content, int* numrowsp, int* numfieldsp)
{
    int stat = OC_NOERR;
    OCnode* node = content->node;
    size_t numrows;

    if(node->octype != OC_Sequence) return THROW(OC_EINVAL);
    *numfieldsp = oclistlength(node->subnodes);
    numrows = oc_recordcount(content->state,content);
    if(numrowsp != NULL)
	*numrowsp = numrows;
    return THROW(stat);
}


/*
opendap_get_sequence_field()
   argv[0] - (Input ) node pointer.
   argv[1] - (Input ) content pointer.
   argv[2] - (Input ) field number
   argv[3] - (Output) Named IDL variable to receive field name
   argv[4] - (Output) Named IDL variable to receive field type
   argv[5] - (Output) Named IDL variable to receive field sub node
   argv[6] - (Output) Named IDL variable to receive field content pointer
   returns - Status (1=ok, 0=fail)
*/

/*
The following two interface procedures present a problem
because they assume that one gets the field before getting
the row -- this is consistent with the way ocapi stores
its sequence data (by column, not row as in the xdr data).
So, the fieldcontent returned is marked as which field,
but its xdroffset remains at the whole sequence.
This fact is used in opendap_get_sequence_row to extract the
correct row X field content.
*/

int
opendap_get_sequence_field(OCnode* node, OCcontent* content, int ifield,
			   char** fieldnamep, OCtype* fieldtypep,
			   OCnode** fieldnodep, OCcontent** fieldcontentp)
{
    OCcontent* fieldcontent;
    OCnode* fieldnode;
    OCstate* state;

    if(node != content->node) return THROW(OC_EINVAL);
    if(node->octype != OC_Sequence) return THROW(OC_EINVAL);

    if(ifield < 0 || ifield >= oclistlength(node->subnodes)) return THROW(OC_EINVAL);
    fieldnode = (OCnode*)oclistget(node->subnodes,ifield);

    state = node->owner;

    /* fieldcontent will be copy of content, but with fieldindex set*/
    fieldcontent = Imakecontent(state,node);
    fieldcontent->mode = content->mode; /* same mode*/
    fieldcontent->index = content->index;  /* same record*/

    /* checkpoint this location as same as oclist.content*/
    fieldcontent->xdroffset = content->xdroffset;

    /* Track the fact that this procedure was called*/
    Isetfieldindex(state,fieldcontent,ifield);

    /* set return values*/
    *fieldnamep = strdup(fieldnode->name);
    *fieldtypep = fieldnode->octype;
    *fieldnodep = fieldnode;
    *fieldcontentp = fieldcontent;

    return THROW(OC_NOERR);
}


/*
opendap_get_sequence_row()
Get the pointer from the sequence aFieldValue[] for the given row
i.e. get ith column from the content
   argv[0] - (Input ) node pointer.
   argv[1] - (Input ) content pointer.
   argv[2] - (Input ) field type
   argv[3] - (Input ) row number
   argv[4] - (Output) Named IDL variable to receive row value ptr
   returns - Status (1=ok, 0=fail)
*/

int
opendap_get_sequence_row(OCnode* fieldnode, OCcontent* fieldcontent, OCtype fieldtype,
                         int irow, OCcontent** rowfieldcontentp)
{
    int stat = OC_NOERR;
    OCstate* state = fieldcontent->state;
    OCnode* seqnode;
    OCcontent* seqcontent;
    OCcontent* rowcontent;
    OCcontent* rowfieldcontent;
    XDR* xdrs;
    Icontent* ic;

    /* Remember that the fieldcontent is really clone of parent oclist.content*/
    seqcontent = fieldcontent;
    seqnode = seqcontent->node;
    /* do some sanity checks*/
    if(seqcontent->mode != Recordmode) return THROW(OC_EINVAL);
    if(seqnode->octype != OC_Sequence) return THROW(OC_EINVAL);

    ic = (Icontent*)fieldcontent->public;
    if(ic == NULL) return THROW(OC_EINVALCOORDS);

    if(findfield(seqnode,fieldnode) != ic->fieldindex) return THROW(OC_EINVAL);
    if(fieldtype != fieldnode->octype) return THROW(OC_EINVAL);

    xdrs = state->dap.xdrs;
    if(xdrs == NULL) return THROW(OC_EXDR);

    /* Move to the specific row of the sequence*/
    seqcontent->index = irow;
    rowcontent = Ilocatecontent(state,seqnode);
    if(rowcontent == NULL) return THROW(stat);

    /* move to specific field*/
if(0) {
    rowfieldcontent = Imakecontent(state,fieldnode);
    stat = oc_fieldcontent(state,rowcontent,rowfieldcontent,ic->fieldindex);
} else {
    rowfieldcontent = Ilocatecontent0(state,rowcontent,fieldnode);
}
    oc_freecontent(state,rowcontent); /* no longer needed*/

    if(stat != OC_NOERR) return THROW(stat);

    if(rowfieldcontentp != NULL) *rowfieldcontentp = rowfieldcontent;

    return THROW(stat);
}

static int
consistent(OCnode* node)
{
    int i;
    if(node->octype == OC_Attribute) return 1;
    if(node->octype == OC_Attributeset) return 1;
    if(node->array.rank > 0 && node->octype != OC_Primitive)
	return 0;
    for(i=0;i<oclistlength(node->subnodes);i++) {
	OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	if(!consistent(subnode)) return 0;
    }
    return 1;
}

/* Result content is guaranteed to be new content instance*/
static OCcontent*
Ilocatecontent(OCstate* state, OCnode* node)
{
    OCcontent* parentcontent;
    OCcontent* content;
    OCnode* parent;

    /* First try: see if node has associated content*/
    content = (OCcontent*)node->public;
    if(content != NULL) return oc_clonecontent(state,content);

    /* note: root must have content*/
    if(node->octype == OC_Dataset) return NULL;

    /* Second try: recursively get parent content*/
    /* Assumes that the parent nodes have requisite*/
    /* current content.*/

    assert(node->octype != OC_Dataset);
    parent = node->container;
    parentcontent = Ilocatecontent(state,parent);
    if(parentcontent == NULL) return NULL; /* missing content somewhere*/

    return Ilocatecontent0(state,parentcontent,node);

}

static OCcontent*
Ilocatecontent0(OCstate* state, OCcontent* parentcontent, OCnode* node)
{
    int stat = OC_NOERR;
    OCcontent* content;
    OCnode* subnode;
    OCnode* parent;
    int fieldindex;
    OCcontent* tmp;

    parent = parentcontent->node;

    /* Compute node's content from parent content*/
    content = Imakecontent(state,node);

    switch (node->octype) {

    case OC_Primitive:
    case OC_Structure:
    case OC_Grid:
    case OC_Sequence:
	fieldindex = findfield(parent,node);
	if(parent->octype == OC_Sequence) {
	    tmp = Imakecontent(state,node);
	    /* We have to pick a record to go thru; use parent specified*/
	    stat = oc_recordcontent(state,parentcontent,tmp,parentcontent->index);
	    if(stat != OC_NOERR) return NULL;
	    /* pick the field*/
	    stat = oc_fieldcontent(state,tmp,content,fieldindex);
	    oc_freecontent(state,tmp);
	    if(stat != OC_NOERR) return NULL;
	} else {
	    /* pick the field*/
	    stat = oc_fieldcontent(state,parentcontent,content,fieldindex);
	    if(stat != OC_NOERR) return NULL;
	}
        break;

    case OC_Dataset: /* Should never happen*/
    default: OCPANIC1("idl-locatecontent: unexpected node type: %d",subnode->octype);
    }
    return content;
}

static int
findfield(OCnode* parent, OCnode* node)
{
    int i;
    for(i=0;i<oclistlength(parent->subnodes);i++) {
	OCnode* subnode = (OCnode*)oclistget(parent->subnodes,i);
	if(node == subnode) return i;
    }
    return 0;
}
